# -*- coding: utf-8 -*-
# Copyright (C) 2022-2024 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause

# Copyright (C) 2022 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause


class CondaProjectError(Exception):
    pass


class CondaProjectLockFailed(CondaProjectError):
    pass


class CommandNotFoundError(CondaProjectError):
    pass



# -*- coding: utf-8 -*-
# Copyright (C) 2022-2024 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause

# Copyright (C) 2022 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import json
import os
import shlex
import signal
import subprocess
from contextlib import nullcontext
from functools import lru_cache
from logging import Logger
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, NoReturn, Optional, Tuple, Union, cast

import conda_lock._vendor.conda.gateways.logging  # noqa: F401
import pexpect
from conda_lock._vendor.conda.core.prefix_data import PrefixData, PrefixRecord
from conda_lock._vendor.conda.models.channel import Channel
from conda_lock._vendor.conda.utils import wrap_subprocess_call
from conda_lock.conda_lock import (
    default_virtual_package_repodata,
    make_lock_spec,
    parse_conda_lock_file,
)
from conda_lock.lockfile.v2prelim.models import Lockfile

from .exceptions import CondaProjectError
from .project_file import EnvironmentYaml, UniqueOrderedList
from .utils import Spinner, detect_shell, execvped, is_windows

CONDA_EXE = os.environ.get("CONDA_EXE", "conda")
CONDA_ROOT = os.environ.get("CONDA_ROOT")
CONDA_PREFIX = os.environ.get("CONDA_PREFIX")


def call_conda(
    args: List[str],
    condarc_path: Optional[Path] = None,
    verbose: bool = False,
    logger: Optional[Logger] = None,
    variables: Optional[dict[str, str]] = None,
) -> subprocess.CompletedProcess:
    """Call conda CLI with subprocess.run"""

    parent_process_env = os.environ.copy()

    variables = {} if variables is None else variables
    env = {**variables, **parent_process_env}

    if condarc_path is not None:
        if logger is not None:
            logger.info(f"setting CONDARC env variable to {condarc_path}")
        env["CONDARC"] = str(condarc_path)

    cmd = [CONDA_EXE] + args

    if verbose:
        stdout = None
    else:
        stdout = subprocess.PIPE

    if logger is not None:
        logger.info(f'running conda command: {" ".join(cmd)}')

    proc = subprocess.run(
        cmd, env=env, stdout=stdout, stderr=subprocess.PIPE, encoding="utf-8"
    )

    if proc.returncode != 0:
        print_cmd = " ".join(cmd)
        raise CondaProjectError(f"Failed to run:\n  {print_cmd}\n{proc.stderr.strip()}")

    return proc


def is_conda_env(prefix: Path) -> bool:
    return (prefix / "conda-meta" / "history").exists()


def conda_prefix(env: Optional[Union[str, Path]] = None) -> Path:
    """Return the path to a conda environment"""

    if env is None:
        return Path(os.environ["CONDA_PREFIX"]).resolve()

    elif env in ("base", "root"):
        return Path(os.environ["CONDA_ROOT"]).resolve()

    else:
        env = Path(env) if isinstance(env, str) else env

        if is_conda_env(env):
            return env.resolve()

        else:
            for d in conda_info()["envs_dirs"]:
                p = Path(d) / env
                if is_conda_env(p):
                    return p.resolve()
            else:
                raise ValueError(f"{env} is not a valid conda environment")


def env_export(
    prefix: Path,
    from_history: bool = True,
    pin_versions: bool = True,
    verbose: bool = False,
) -> Tuple[EnvironmentYaml, Lockfile]:
    """Create an environment.yml spec and lockfile from an existing environment"""

    pd = PrefixData(prefix, pip_interop_enabled=True)
    pkgs = cast(List[PrefixRecord], pd.iter_records())

    channels: List[Channel] = []
    dependencies: List[Union[str, Dict[str, List[str]]]] = []
    pip: List[str] = []

    n_conda = 0
    n_pip = 0
    for p in pkgs:
        if p.schannel == "pypi":
            n_pip += 1
            if from_history:
                for fn in p.files:
                    fn = prefix / Path(fn)
                    if fn.name == "REQUESTED":
                        pip.append(f"{p.name}=={p.version}")
                        break
            else:
                pip.append(f"{p.name}=={p.version}")
            continue

        n_conda += 1

        if from_history and p.requested_spec == "None":
            continue

        if from_history and pin_versions:
            spec = f"{p.schannel}::{p.name}={p.version}"
        elif from_history and (not pin_versions):
            spec = f"{p.schannel}::{p.requested_spec}"
        else:
            spec = f"{p.schannel}::{p.name}={p.version}={p.build}"
        dependencies.append(spec)

        channels.append(p.channel)

    full_export = (not from_history) or (
        len(pip) + len(dependencies) == n_conda + n_pip
    )
    empty = n_conda + n_pip == 0

    if pip:
        dependencies.append({"pip": pip})

    if full_export:
        subdirs = set(c.subdir for c in channels)
        subdirs.discard("pypi")
        subdirs.discard("noarch")
        if not subdirs:
            platforms = [current_platform()]
        else:
            platforms = list(subdirs)
    else:
        platforms = DEFAULT_PLATFORMS

    channel_names = UniqueOrderedList(
        [Channel.from_url(url).canonical_name for url in conda_info()["channels"]]
    )

    environment = EnvironmentYaml(
        name=prefix.name,
        channels=channel_names,
        dependencies=dependencies,
        platforms=DEFAULT_PLATFORMS if empty else platforms,
        prefix=prefix,
    )
    with Spinner("Constructing lockfile") if verbose else nullcontext():
        with TemporaryDirectory() as tmp:
            tmp = Path(tmp)

            requested = tmp / "requested.yml"
            environment.yaml(requested)
            spec = make_lock_spec(
                src_files=[requested],
                virtual_package_repo=default_virtual_package_repodata(),
            )

            if full_export or empty:
                exported = requested
            else:
                exported = tmp / "exported.yml"
                call_conda(
                    ["env", "export", "-p", str(prefix), "--file", str(exported)]
                )

            lock = tmp / "conda-lock.yml"
            call_conda(
                [
                    "lock",
                    "-f",
                    str(exported),
                    "--lockfile",
                    str(lock),
                    "-p",
                    current_platform(),
                    "--metadata",
                    "timestamp",
                ]
            )
            lock_content = parse_conda_lock_file(lock)
            lock_content.metadata.content_hash = spec.content_hash()
            lock_content.metadata.sources = ["environment.yml"]

    return environment, lock_content


def conda_info():
    proc = call_conda(["info", "--json"])
    parsed = json.loads(proc.stdout)
    return parsed


@lru_cache()
def current_platform() -> str:
    """Load the current platform by calling conda info."""
    info = conda_info()
    return info.get("platform")


DEFAULT_PLATFORMS = set(["osx-64", "win-64", "linux-64", current_platform()])


def conda_run(
    cmd: str,
    prefix: Path,
    working_dir: Path,
    env: Optional[Dict[str, str]] = None,
    extra_args: Optional[List[str]] = None,
) -> NoReturn:
    extra_args = [] if extra_args is None else extra_args
    arguments = shlex.split(cmd) + extra_args

    _, (shell, *args) = wrap_subprocess_call(
        root_prefix=CONDA_ROOT,
        prefix=str(prefix),
        dev_mode=False,
        debug_wrapper_scripts=False,
        arguments=arguments,
        use_system_tmp_path=True,
    )

    env = {} if env is None else env

    if not is_windows():
        args = ["-c", *args]

    execvped(file=shell, args=args, env=env, cwd=working_dir)


def _send_activation(child_shell: pexpect.spawn, prefix):
    def sigwinch_passthrough(sig, data):
        if not child_shell.closed:
            t = os.get_terminal_size()
            child_shell.setwinsize(t.lines, t.columns)

    t = os.get_terminal_size()
    child_shell.setwinsize(t.lines, t.columns)
    signal.signal(signal.SIGWINCH, sigwinch_passthrough)
    child_shell.sendline(f"conda activate {prefix}")
    child_shell.interact()
    child_shell.close()


def conda_activate(prefix: Path, working_dir: Path, env: Optional[Dict] = None):
    env = {} if env is None else env

    shell_path, shell_name = detect_shell()

    args = []
    if is_windows():
        if shell_name in ["powershell", "pwsh"]:
            conda_hook = str(Path(CONDA_ROOT) / "shell" / "condabin" / "conda-hook.ps1")
            args = [
                "-ExecutionPolicy",
                "ByPass",
                "-NoExit",
                conda_hook,
                ";",
                "conda",
                "activate",
                str(prefix),
            ]
        elif shell_name == "cmd":
            activate_bat = str(Path(CONDA_ROOT) / "Scripts" / "activate.bat")
            args = ["/K", activate_bat, str(prefix)]
    else:
        args = ["-i"]

    activate_message = (
        f"## Project environment {prefix.name} activated in a new shell.\n"
        f"## Exit this shell to de-activate."
    )
    print(activate_message)

    if is_windows():
        subprocess.run([shell_path, *args], cwd=working_dir, env=env)
    else:
        child_shell = pexpect.spawn(
            command=shell_path, args=args, cwd=working_dir, env=env, echo=True
        )

        _send_activation(child_shell, prefix)



# Copyright (C) 2022-2024 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause

# Copyright (C) 2022 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause

import itertools
import os
import platform
import sys
import threading
import time
from collections import ChainMap
from collections.abc import Generator
from contextlib import contextmanager
from inspect import Traceback
from itertools import groupby
from pathlib import Path
from subprocess import Popen
from typing import Callable, Dict, List, NoReturn, Optional, Type, Union

import shellingham
from dotenv import dotenv_values

from .exceptions import CondaProjectError


@contextmanager
def env_variable(key: str, value: str) -> Generator:
    """Temporarily set environment variable in a context manager."""
    old = os.environ.get(key, None)
    os.environ[key] = value

    yield

    if old is None:
        os.environ.pop(key, None)
    else:
        os.environ[key] = old


class Spinner:
    """Multithreaded CLI spinner context manager

    Attributes:
        prefix: Text to display at the start of the line

    Args:
        prefix: Text to display at the start of the line

    """

    def __init__(self, prefix: str):
        self.prefix = prefix
        self._event = threading.Event()
        self._thread = threading.Thread(target=self._spin)

    def _spin(self) -> None:
        spinner = itertools.cycle(["◜", "◠", "◝", "◞", "◡", "◟"])

        while not self._event.is_set():
            sys.stdout.write("\r")
            sys.stdout.write("\033[K")
            sys.stdout.write(f"{self.prefix}: {next(spinner)} ")
            sys.stdout.flush()
            time.sleep(0.10)

    def start(self) -> None:
        self._thread.start()

    def stop(self) -> None:
        self._event.set()
        self._thread.join()
        sys.stdout.write("\r")
        sys.stdout.write("\033[K")
        sys.stdout.write(f"{self.prefix}: done\n")
        sys.stdout.flush()

    def __enter__(self) -> None:
        self.start()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_tb: Optional[Traceback],
    ) -> None:
        self.stop()


def find_file(directory: Path, options: tuple) -> Optional[Path]:
    """Search for a file in a directory from a tuple of variants.

    Returns:
        The path to the file if found else None

    Raises:
        CondaProjectError if more than one of the options is found

    """
    found = []

    for filename in options:
        path = directory / filename
        if path.exists():
            found.append(path.resolve())

    if len(found) == 1:
        return found[0]
    elif len(found) > 1:
        _found_files = "\n".join([str(p) for p in found])
        raise CondaProjectError(
            f"Multiple variants of the same file were found.\n{_found_files}\nConsider using one of them."
        )
    else:
        return None


def merge_dicts(*dicts):
    return dict(ChainMap(*reversed(dicts)))


def prepare_variables(project_directory: Path, *variable_dicts) -> Dict[str, str]:
    variables = [{} if vars is None else vars for vars in variable_dicts]

    dotenv = dotenv_values(project_directory / ".env")

    env = merge_dicts(*variables, dotenv, os.environ)

    missing_vars = [k for k, v in env.items() if v is None]
    if missing_vars:
        errs = "\n".join(missing_vars)
        msg = (
            "The following variables do not have a default value and values\n"
            "were not provided in the .env file or set on the command line"
            f" when executing 'conda project run':\n{errs}"
        )
        raise CondaProjectError(msg)

    return env


def is_windows():
    return platform.system() == "Windows"


def execvped(
    file: Union[Path, str], args: List[str], env: Dict[str, str], cwd: Union[Path, str]
) -> NoReturn:
    """A cross-platform os.execvpe - like executor

    The goal is the be able to launch a command in a working directory,
    with environment variables and exit to the shell with the return code
    of the command and ensure that on error the previous working directory
    is restored.

    The "d" in the function name refers to the requirement that
    the working directory (cwd) flag be used.
    """

    sys.stdout.flush()
    sys.stderr.flush()

    if is_windows():
        sys.exit(Popen(args=[file, *args], env=env, cwd=cwd).wait())
    else:
        old_dir = Path.cwd()
        try:
            os.chdir(cwd)
            os.execvpe(file, args, env)
        finally:
            os.chdir(old_dir)


def detect_shell():
    try:
        shell_name, shell_path = shellingham.detect_shell()
    except shellingham.ShellDetectionFailure:
        if os.name == "posix":
            shell_name = shell_path = os.environ.get("SHELL", "/bin/sh")
        elif os.name == "nt":
            shell_name = shell_path = os.environ.get("COMSPEC", "cmd.exe")
        else:
            raise RuntimeError(
                "Could not determine an appropriate shell to activate for your OS."
            )

    return shell_name, shell_path


def dedupe_list_of_dicts(data: list, key: Callable, keep: Callable) -> list:
    deduped = []
    for _, g in groupby(sorted(data, key=key), key=key):
        values = list(g)
        if len(values) > 1:
            deduped.extend(list(filter(keep, values)))
        else:
            deduped.extend(values)
    return deduped


def get_envs_paths() -> List[Path]:
    specified_path = os.environ.get("CONDA_PROJECT_ENVS_PATH", "")
    env_paths = specified_path.split(os.pathsep) if specified_path else []
    expanded_paths = [Path(os.path.expandvars(path)) for path in env_paths]
    return expanded_paths



# Copyright (C) 2022-2024 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause

# Copyright (C) 2024 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import json
from collections.abc import Iterable
from pathlib import Path
from typing import Any, Dict, List, Optional, OrderedDict, TextIO, Union

from conda_lock._vendor.conda.models.match_spec import MatchSpec
from pkg_resources import Requirement
from ruamel.yaml import YAML

try:  # pragma: no cover
    # Version 2 provides a v1 API
    from pydantic.v1 import BaseModel, ValidationError, validator  # pragma: no cover
except ImportError:  # pragma: no cover
    from pydantic import BaseModel  # type: ignore; #pragma: no cover
    from pydantic import ValidationError  # type: ignore; #pragma: no cover
    from pydantic import validator  # type: ignore; #pragma: no cover

from .exceptions import CondaProjectError

PROJECT_YAML_FILENAMES = ("conda-project.yml", "conda-project.yaml")
ENVIRONMENT_YAML_FILENAMES = ("environment.yml", "environment.yaml")

yaml = YAML(typ="rt")
yaml.default_flow_style = False
yaml.block_seq_indent = 2
yaml.indent = 2


def _cleandict(d: Dict) -> Dict:
    return {k: v for k, v in d.items() if v is not None}


class BaseYaml(BaseModel):
    def yaml(self, stream: Union[TextIO, Path], drop_empty_keys=False):
        # Passing through self.json() allows json_encoders
        # to serialize objects.
        object_hook = _cleandict if drop_empty_keys else None
        encoded = json.loads(self.json(), object_hook=object_hook)
        return yaml.dump(encoded, stream)

    @classmethod
    def parse_yaml(cls, fn: Union[str, Path]):
        d = yaml.load(fn)
        if d is None:
            msg = (
                f"Failed to read {fn} as {cls.__name__}. The file appears to be empty."
            )
            raise CondaProjectError(msg)
        try:
            return cls(**d)
        except ValidationError as e:
            msg = f"Failed to read {fn} as {cls.__name__}\n{str(e)}"
            raise CondaProjectError(msg)

    class Config:
        json_encoders = {Path: lambda v: v.as_posix()}


class Command(BaseModel):
    cmd: str
    environment: Optional[str] = None
    variables: Optional[Dict[str, Optional[str]]] = None

    class Config:
        extra = "forbid"


class CondaProjectYaml(BaseYaml):
    name: str
    environments: OrderedDict[str, List[Path]]
    variables: Dict[str, Optional[str]] = {}
    commands: OrderedDict[str, Union[Command, str]] = OrderedDict()


class UniqueOrderedList(list):
    def __init__(self, iterable: Iterable):
        uniques = []
        for item in iterable:
            if item not in uniques:
                uniques.append(item)

        super().__init__(uniques)

    def _remove_duplicates(self, other: Any):
        for item in self:
            if item in other:
                other.remove(item)

    def append(self, __object: Any) -> None:
        if __object not in self:
            return super().append(__object)

    def extend(self, __iterable: Iterable) -> None:
        self._remove_duplicates(__iterable)
        return super().extend(__iterable)


class EnvironmentYaml(BaseYaml):
    name: Optional[str] = None
    channels: Optional[Union[UniqueOrderedList, List[str]]] = None
    dependencies: List[Union[str, Dict[str, List[str]]]] = []
    variables: Optional[Dict[str, str]] = None
    prefix: Optional[Path] = None
    platforms: Optional[List[str]] = None

    @validator("dependencies")
    def only_pip_key_allowed(cls, v):
        for item in v:
            if isinstance(item, dict):
                if not item.keys() == {"pip"}:
                    raise ValueError(
                        f'The dependencies key contains an invalid map {item}. Only "pip:" is allowed.'
                    )
        return v

    @validator("channels")
    def convert_channels_list(cls, v):
        if not isinstance(v, UniqueOrderedList):
            return UniqueOrderedList(v)
        else:
            return v

    @property
    def conda_matchspecs(self):
        return [
            MatchSpec(dep) for dep in self.dependencies if not isinstance(dep, dict)
        ]

    @property
    def _pip_requirements(self):
        return [d for d in self.dependencies if isinstance(d, dict) and "pip" in d]

    @property
    def pip_requirements(self):
        pip = self._pip_requirements
        if pip:
            return [Requirement(dep) for dep in self._pip_requirements[0]["pip"]]
        else:
            return []

    def _add_pip_requirements(self, reqs):
        if "pip" not in self.dependencies:
            print(
                "Warning: you have pip-installed dependencies in your environment file, "
                "but you do not list pip itself as one of your conda dependencies. Please "
                "add an explicit pip dependency. I'm adding one for you, but still nagging you."
            )
            self.dependencies.append("pip")
        pip = self._pip_requirements
        if pip:
            pip[0]["pip"].extend(reqs)
        else:
            self.dependencies.append({"pip": reqs})

    def _replace_pip_requirement(self, idx, req):
        pip = self._pip_requirements
        if pip:
            pip[0]["pip"][idx] = req

    def _remove_pip_requirement(self, idx):
        pip = self._pip_requirements
        if pip:
            pip[0]["pip"].pop(idx)

    def add_dependencies(
        self,
        dependencies: List[str],
        channels: Optional[Union[UniqueOrderedList, List[str]]] = None,
    ) -> None:
        current_conda_names = [dep.name for dep in self.conda_matchspecs]
        current_pip_names = [dep.name for dep in self.pip_requirements]

        conda_to_add = []
        pip_to_add = []
        for dep in dependencies:
            if dep.startswith("@pip::"):
                _, dep = dep.split("::", maxsplit=1)
                name = Requirement(dep).name
                if name in current_pip_names:
                    self._replace_pip_requirement(current_pip_names.index(name), dep)
                else:
                    pip_to_add.append(dep)
            else:
                name = MatchSpec(dep).name
                if name in current_conda_names:
                    self.dependencies[current_conda_names.index(name)] = dep
                else:
                    conda_to_add.append(dep)

        self.dependencies.extend(conda_to_add)
        if pip_to_add:
            self._add_pip_requirements(pip_to_add)

        if channels:
            if self.channels:
                self.channels.extend(channels)
            else:
                self.channels = UniqueOrderedList(channels)

    def remove_dependencies(self, dependencies: List[str]) -> None:
        current_conda_names = [dep.name for dep in self.conda_matchspecs]
        current_pip_names = [dep.name for dep in self.pip_requirements]

        for dep in dependencies:
            if dep.startswith("@pip::"):
                _, dep = dep.split("::", maxsplit=1)
                name = Requirement(dep).name
                if name in current_pip_names:
                    self._remove_pip_requirement(current_pip_names.index(name))
            else:
                name = MatchSpec(dep).name
                if name in current_conda_names:
                    self.dependencies.pop(current_conda_names.index(name))
                    current_conda_names.remove(name)



# -*- coding: utf-8 -*-
# Copyright (C) 2022-2024 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause

# Copyright (C) 2022 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import typing
from argparse import REMAINDER, ArgumentParser

from conda_project import __version__

from ..project import DEFAULT_PLATFORMS
from . import commands

if typing.TYPE_CHECKING:
    # This is here to prevent potential future breaking API changes
    # in argparse from affecting at runtime
    from argparse import _SubParsersAction


def cli() -> ArgumentParser:
    """Construct the command-line argument parser."""
    common = ArgumentParser(add_help=False)
    common.add_argument(
        "--directory",
        metavar="PROJECT_DIR",
        default=".",
        help="Project directory (defaults to current directory)",
    )

    extras = ArgumentParser(add_help=False)
    extras.add_argument(
        "--project-archive",
        metavar="PROJECT_ARCHIVE_FILE_OR_URL",
        help=(
            "EXPERIMENTAL: Extract and run directly from a project archive. The archive can be a local file "
            "or a fsspec compatible URL. You may need to install appropriate driver packages to work with "
            "remote archives. Optionally, use --directory to set the destination directory of the extracted "
            "project."
        ),
    )
    extras.add_argument(
        "--archive-storage-options",
        help=(
            "EXPERIMENTAL: Comma separated list of fsspec storage_options for accessing a remote archive "
            "For example --archive-storage-options username=<user>,password=<pass>"
        ),
        action="store",
        default=None,
    )

    p = ArgumentParser(
        description="Tool for encapsulating, running, and reproducing projects with conda environments",
        conflict_handler="resolve",
    )
    p.add_argument(
        "-V",
        "--version",
        action="version",
        help="Show the conda-prefix-replacement version number and exit.",
        version="conda_project %s" % __version__,
    )

    subparsers = p.add_subparsers(metavar="command", required=True)

    _create_init_parser(subparsers, common)
    _create_lock_parser(subparsers, common, extras)
    _create_check_parser(subparsers, common, extras)
    _create_install_parser(subparsers, common, extras)
    _create_add_parser(subparsers, common, extras)
    _create_remove_parser(subparsers, common, extras)
    _create_activate_parser(subparsers, common, extras)
    _create_clean_parser(subparsers, common)
    _create_run_parser(subparsers, common, extras)

    return p


def _create_init_parser(
    subparsers: "_SubParsersAction", *parent_parsers: ArgumentParser
) -> None:
    """Add a subparser for the "init" and "create" subcommands.

    Args:
        subparsers: The existing subparsers corresponding to the "command" meta-variable.
        parent_parsers: The parent parsers, which are used to pass common arguments into the subcommands.

    """
    desc = "Initialize a new project"

    # TODO: If we deprecate "create", this loop can go away.
    for subcommand_name in ["init", "create"]:
        p = subparsers.add_parser(
            subcommand_name, description=desc, help=desc, parents=parent_parsers
        )
        p.add_argument(
            "-n", "--name", help="Name for the project.", action="store", default=None
        )
        p.add_argument(
            "-c",
            "--channel",
            help=(
                "Additional channel to search for packages. The default channel is 'defaults'. "
                "Multiple channels are added with repeated use of this argument."
            ),
            action="append",
        )
        p.add_argument(
            "--platforms",
            help=(
                f"Comma separated list of platforms for which to lock dependencies. "
                f"The default is {','.join(DEFAULT_PLATFORMS)}"
            ),
            action="store",
            default=",".join(DEFAULT_PLATFORMS),
        )
        p.add_argument(
            "--conda-configs",
            help=(
                "Comma separated list of conda configuration parameters to write into the "
                ".condarc file in the project directory. The format for each config is key=value. "
                "For example --conda-configs experimental_solver=libmamba,channel_priority=strict"
            ),
            action="store",
            default=None,
        )
        p.add_argument(
            "--lock",
            help="Create the conda-lock.<env>.yml file(s)",
            action="store_true",
        )
        p.add_argument(
            "--install",
            help="Create the local conda environment for the current platform.",
            action="store_true",
        )
        p.add_argument(
            "--from-environment",
            help=(
                "Initialize the default environment spec and lock from an existing"
                "conda environment by name or prefix."
            ),
            action="store",
            default=None,
        )
        p.add_argument(
            "dependencies",
            help=(
                "Packages to add to the environment.yml. The format for each package is '<name>[<op><version>]' "
                "where <op> can be =, <, >, <=, or >=."
            ),
            action="store",
            nargs="*",
            metavar="PACKAGE_SPECIFICATION",
        )

        p.set_defaults(func=getattr(commands, subcommand_name))


def _create_lock_parser(
    subparsers: "_SubParsersAction", *parent_parsers: ArgumentParser
) -> None:
    """Add a subparser for the "lock" subcommand.

    Args:
        subparsers: The existing subparsers corresponding to the "command" meta-variable.
        parent_parsers: The parent parsers, which are used to pass common arguments into the subcommands.

    """
    desc = "Lock all conda environments or a specific one by creating conda-lock.<env>.yml file(s)."

    p = subparsers.add_parser(
        "lock", description=desc, help=desc, parents=parent_parsers
    )
    p.add_argument(
        "environment",
        help="Optional: Lock the selected environment. If no environment name is selected "
        "all environments are locked.",
        nargs="?",
    )
    p.add_argument(
        "--force",
        help="Remove and recreate existing conda-lock.<env>.yml file(s).",
        action="store_true",
    )

    p.set_defaults(func=commands.lock)


def _create_check_parser(
    subparsers: "_SubParsersAction", *parent_parsers: ArgumentParser
) -> None:
    """Add a subparser for the "lock" subcommand.

    Args:
        subparsers: The existing subparsers corresponding to the "command" meta-variable.
        parent_parsers: The parent parsers, which are used to pass common arguments into the subcommands.

    """
    desc = (
        "Check the project for inconsistencies or errors. This will check that conda-lock.<env>.yml file(s) "
        "have been created for each environment and are up-to-date with the source environment specifications. "
        "If the project is fully locked this command will not print anything and return status code 0. If any "
        "environment is not fully locked details are printed to stderr and the command returns status code 1."
    )

    p = subparsers.add_parser(
        "check", description=desc, help=desc, parents=parent_parsers
    )

    p.set_defaults(func=commands.check)


def _create_install_parser(
    subparsers: "_SubParsersAction", *parent_parsers: ArgumentParser
) -> None:
    """Add a subparser for the "install" and "prepare" subcommands.

    Args:
        subparsers: The existing subparsers corresponding to the "command" meta-variable.
        parent_parsers: The parent parsers, which are used to pass common arguments into the subcommands.

    """
    desc = "Install the packages into the conda environments"

    for subcommand_name in ["install", "prepare"]:
        p = subparsers.add_parser(
            subcommand_name, description=desc, help=desc, parents=parent_parsers
        )
        group = p.add_mutually_exclusive_group(required=False)
        group.add_argument(
            "environment",
            help="Prepare the selected environment. If no environment name is selected "
            "the first environment defined in the conda-project.yml file is prepared.",
            nargs="?",
        )
        group.add_argument(
            "--as-platform",
            help="Prepare the conda environment assuming a different platform/subdir name.",
            action="store",
            metavar="PLATFORM",
        )
        group.add_argument(
            "--all",
            help="Check or prepare all defined environments.",
            action="store_true",
        )
        p.add_argument(
            "--check-only",
            help="Check that the prepared conda environment exists and is up-to-date with the "
            "source environment and files and lockfile and then exit. If the environment is up-to-date "
            "nothing is printed and the command exists with 0. If the environment is missing or out-of-date "
            "details are printed to stderr and the command exits with 1.",
            action="store_true",
        )
        p.add_argument(
            "--force",
            help="Remove and recreate an existing environment.",
            action="store_true",
        )

        p.set_defaults(func=getattr(commands, subcommand_name))


def _create_add_parser(
    subparsers: "_SubParsersAction", *parent_parsers: ArgumentParser
) -> None:
    """Add a subparser for the "add" subcommand.

    Args:
        subparsers: The existing subparsers corresponding to the "command" meta-variable.
        parent_parsers: The parent parsers, which are used to pass common arguments into the subcommands.

    """
    desc = "Add packages to an environment"

    # TODO: If we deprecate "create", this loop can go away.
    p = subparsers.add_parser(
        "add", description=desc, help=desc, parents=parent_parsers
    )
    p.add_argument("--environment", default=None)
    p.add_argument(
        "-c",
        "--channel",
        help=(
            "Additional channel to search for packages. The default channel is 'defaults'. "
            "Multiple channels are added with repeated use of this argument."
        ),
        action="append",
    )
    p.add_argument(
        "dependencies",
        help=(
            "Packages to add to the environment.yml. The format for each package is "
            "'[<prefix>::]<name>[<op><version>]' where <op> can be =, <, >, <=, or >=. "
            "Most commonly `<prefix>::` declares the conda channel from which to install packages. Use the "
            "prefix `@pip::` to add pypi package dependencies with support for full pypi package specification "
            "syntax."
        ),
        action="store",
        nargs="*",
        metavar="PACKAGE_SPECIFICATION",
    )

    p.set_defaults(func=commands.add)


def _create_remove_parser(
    subparsers: "_SubParsersAction", *parent_parsers: ArgumentParser
) -> None:
    """Add a subparser for the "remove" subcommand.

    Args:
        subparsers: The existing subparsers corresponding to the "command" meta-variable.
        parent_parsers: The parent parsers, which are used to pass common arguments into the subcommands.

    """
    desc = "Remove packages to an environment"

    # TODO: If we deprecate "create", this loop can go away.
    p = subparsers.add_parser(
        "remove", description=desc, help=desc, parents=parent_parsers
    )
    p.add_argument("--environment", default=None)
    p.add_argument(
        "dependencies",
        help=(
            "Packages to remove from the environment.yml. Only the name of the package is required here. To remove "
            "a pip package use the pypyi:: prefix."
        ),
        action="store",
        nargs="*",
        metavar="PACKAGE_SPECIFICATION",
    )

    p.set_defaults(func=commands.remove)


def _create_clean_parser(
    subparsers: "_SubParsersAction", *parent_parsers: ArgumentParser
) -> None:
    """Add a subparser for the "clean" subcommand.

    Args:
        subparsers: The existing subparsers corresponding to the "command" meta-variable.
        parent_parsers: The parent parsers, which are used to pass common arguments into the subcommands.

    """
    desc = "Clean the conda environments"

    p = subparsers.add_parser(
        "clean", description=desc, help=desc, parents=parent_parsers
    )
    p.add_argument(
        "environment",
        help="Remove environment prefix for selected environment. If no environment name is selected "
        "the first environment defined in the conda-project.yml file is removed.",
        nargs="?",
    )
    p.add_argument(
        "--all", help="Prepare all defined environments.", action="store_true"
    )

    p.set_defaults(func=commands.clean)


def _create_run_parser(
    subparsers: "_SubParsersAction", *parent_parsers: ArgumentParser
) -> None:
    """Add a subparser for the "run" subcommand.

    Args:
        subparsers: The existing subparsers corresponding to the "command" meta-variable.
        parent_parsers: The parent parsers, which are used to pass common arguments into the subcommands.

    """
    desc = "Run commands in project environments."

    p = subparsers.add_parser(
        "run", description=desc, help=desc, parents=parent_parsers
    )
    p.add_argument(
        "--environment",
        help=(
            "Specify the environment in which to run the command. The default environment for the command is either "
            "what was specified in the command definition or the first environment defined in the project."
        ),
    )
    p.add_argument(
        "--external-environment",
        metavar="ENV_NAME_OR_PREFIX",
        help=(
            "EXPERIMENTAL: Specify the name or prefix path to a conda environment not declared in this project."
        ),
    )
    p.add_argument(
        "command",
        help="Optional: Run a command in the conda environment. The full command can be provided on the CLI "
        "or the name of a command defined in the conda-project.yml file. If no command is provided "
        "the first command in the conda-project.yml file is run if there is one.",
        nargs="?",
    )
    p.add_argument(
        "extra_args",
        help="Optional arguments passed to the command",
        nargs=REMAINDER,
        default=None,
    )

    p.set_defaults(func=commands.run)


def _create_activate_parser(
    subparsers: "_SubParsersAction", *parent_parsers: ArgumentParser
) -> None:
    """Add a subparser for the "run" subcommand.

    Args:
        subparsers: The existing subparsers corresponding to the "command" meta-variable.
        parent_parsers: The parent parsers, which are used to pass common arguments into the subcommands.

    """
    desc = "Activate a conda environment defined in the environment.yml or conda-project.yml files."

    p = subparsers.add_parser(
        "activate", description=desc, help=desc, parents=parent_parsers
    )
    p.add_argument(
        "environment",
        help="Activate selected environment in a new shell. If no environment name is selected "
        "the first environment defined in the conda-project.yml file is activated.",
        nargs="?",
    )

    p.set_defaults(func=commands.activate)


def parse_and_run(args: list[str] | None = None) -> int:
    """Parse the command-line arguments and run the appropriate sub-command.

    Args:
        args: Command-line arguments. Defaults to system arguments.

    Returns:
        The return code to pass to the operating system.

    """
    p = cli()
    parsed_args, _ = p.parse_known_args(args)
    return parsed_args.func(parsed_args)


def main() -> int:
    """Main entry-point into the `conda-project` command-line interface."""
    import sys

    if len(sys.argv) == 1:
        args = ["-h"]
    else:
        args = sys.argv[1:]

    retcode = parse_and_run(args)
    return retcode
